#include "boucane.hpp"
#include "core/paging.hpp"
#include "drivers/memtext.hpp"
#include "core/idt.hpp"
#include "boot/multiboot2.hpp"
#include "drivers/framebuffer.hpp"
#include "drivers/psftext.hpp"
#include "drivers/vgatext.hpp"
#include "libs/stdio.hpp"
#include "libs/string.hpp"

u64 kvar_kernel_vma;
u64 kvar_stack_pma;
u64 kvar_userspace_pma;
u64 kvar_bss_start;
u64 kvar_bss_end;
u64 kvar_terminus_psf_start;
u64 kvar_terminus_psf_end;

void (*printk)(char *str,...)=printf;

extern "C" void boucane(u64 mb_info){
    // Init linker variables
    asm volatile ("movq $__kernel_vma, %0":"=m"(kvar_kernel_vma));
    asm volatile ("movq $__userspace_pma, %0":"=m"(kvar_userspace_pma));
    asm volatile ("movq $__stack_pma, %0":"=m"(kvar_stack_pma));
    asm volatile ("movq $__bss_start, %0":"=m"(kvar_bss_start));
    asm volatile ("movq $__bss_end, %0":"=m"(kvar_bss_end));
    asm volatile ("movq $res_binary_res_terminus_psf_start, %0":"=m"(kvar_terminus_psf_start));
    asm volatile ("movq $res_binary_res_terminus_psf_end, %0":"=m"(kvar_terminus_psf_end));

    // Init data structures
    asm volatile ("call load_gdt");
    paging_enable();
    memtext_init();
    idt_enable_interrupt();

    // Looking for framebuffer
    FRAMEBUFFER fb_info;
    if(mb2_find_framebuffer((u32*)mb_info, &fb_info)){
        if(fb_info.bpp>16){
            FB_CFG conf;
            conf.depth=fb_info.bpp;
            conf.location=fb_info.addr;
            conf.pitch=fb_info.pitch;
            conf.width=fb_info.width;
            conf.height=fb_info.height;
            framebuffer_init(conf);
            psftext_init((void*)kvar_terminus_psf_start);
            __putchar=psftext_putchar;
        }
        else{
            vgatext_init();
            __putchar=vgatext_putchar;
        }
    }

    // Booting!
    printk("Booting Boucane v%d.%d.%d\n",VERSION_MAJOR,VERSION_MINOR, VERSION_PATH);


    while(1);
}